import sys
import random
import time
import pytest
from falcon_caching import Cache
from falcon_caching.cache import function_namespace

from tests.conftest import EVICTION_STRATEGIES


def test_memoize(caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize(3)
    def big_foo(a, b):
        return a + b + random.randrange(0, 100000)

    result = big_foo(5, 2)

    time.sleep(1)

    assert big_foo(5, 2) == result

    result2 = big_foo(5, 3)
    assert result2 != result

    time.sleep(3)

    assert big_foo(5, 2) != result

    time.sleep(1)

    assert big_foo(5, 3) != result2


def test_memoize_hashes(caches, hash_method):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize(3, hash_method=hash_method)
    def big_foo(a, b):
        return a + b + random.randrange(0, 100000)

    result = big_foo(5, 2)

    time.sleep(1)

    assert big_foo(5, 2) == result

    result2 = big_foo(5, 3)
    assert result2 != result

    time.sleep(3)

    assert big_foo(5, 2) != result

    time.sleep(1)

    assert big_foo(5, 3) != result2


def test_memoize_timeout(caches, app):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    cache.cache_options["default_timeout"] = 1

    @cache.memoize()
    def big_foo(a, b):
        return a + b + random.randrange(0, 100000)

    result = big_foo(5, 2)
    assert big_foo(5, 2) == result
    time.sleep(2)
    assert big_foo(5, 2) != result


def test_memoize_annotated(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    if sys.version_info >= (3, 0):
        @cache.memoize(50)
        def big_foo_annotated(a, b):
            return a + b + random.randrange(0, 100000)
        big_foo_annotated.__annotations__ = {'a': int, 'b': int, 'return': int}

        result = big_foo_annotated(5, 2)

        time.sleep(1)

        assert big_foo_annotated(5, 2) == result


def test_memoize_utf8_arguments(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b):
        return "{}-{}".format(a, b)

    big_foo("æøå", "chars")


def test_memoize_unicode_arguments(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b):
        return u"{}-{}".format(a, b)

    big_foo(u"æøå", "chars")


def test_memoize_delete(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize(5)
    def big_foo(a, b):
        return a + b + random.randrange(0, 100000)

    result = big_foo(5, 2)
    result2 = big_foo(5, 3)

    time.sleep(1)

    assert big_foo(5, 2) == result
    assert big_foo(5, 2) == result
    assert big_foo(5, 3) != result
    assert big_foo(5, 3) == result2

    cache.delete_memoized(big_foo)
    cache.delete_memoized(big_foo, 5, 3)
    cache.delete_memoized(big_foo, 5, 2)

    assert big_foo(5, 2) != result
    assert big_foo(5, 3) != result2


def test_memoize_no_timeout_delete(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b):
        return a + b + random.randrange(0, 100000)

    result_a = big_foo(5, 1)
    result_b = big_foo(5, 2)

    assert big_foo(5, 1) == result_a
    assert big_foo(5, 2) == result_b
    cache.delete_memoized(big_foo, 5, 2)

    assert big_foo(5, 1) == result_a
    assert big_foo(5, 2) != result_b

    # Cleanup bigfoo 5,1 5,2 or it might conflict with
    # following run if it also uses memecache
    cache.delete_memoized(big_foo, 5, 2)
    cache.delete_memoized(big_foo, 5, 1)


def test_memoize_verhash_delete(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize(5)
    def big_foo(a, b):
        return a + b + random.randrange(0, 100000)

    result = big_foo(5, 2)
    result2 = big_foo(5, 3)

    time.sleep(1)

    assert big_foo(5, 2) == result
    assert big_foo(5, 2) == result
    assert big_foo(5, 3) != result
    assert big_foo(5, 3) == result2

    cache.delete_memoized_verhash(big_foo)

    _fname, _fname_instance = function_namespace(big_foo)
    version_key = cache._memvname(_fname)
    assert cache.get(version_key) is None

    assert big_foo(5, 2) != result
    assert big_foo(5, 3) != result2

    assert cache.get(version_key) is not None


def test_memoize_annotated_delete(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize(5)
    def big_foo_annotated(a, b):
        return a + b + random.randrange(0, 100000)

    big_foo_annotated.__annotations__ = {'a': int, 'b': int, 'return': int}

    result = big_foo_annotated(5, 2)
    result2 = big_foo_annotated(5, 3)

    time.sleep(1)

    assert big_foo_annotated(5, 2) == result
    assert big_foo_annotated(5, 2) == result
    assert big_foo_annotated(5, 3) != result
    assert big_foo_annotated(5, 3) == result2

    cache.delete_memoized_verhash(big_foo_annotated)

    _fname, _fname_instance = function_namespace(big_foo_annotated)
    version_key = cache._memvname(_fname)
    assert cache.get(version_key) is None

    assert big_foo_annotated(5, 2) != result
    assert big_foo_annotated(5, 3) != result2

    assert cache.get(version_key) is not None


def test_memoize_args(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b):
        return sum(a) + sum(b) + random.randrange(0, 100000)

    result_a = big_foo([5, 3, 2], [1])
    result_b = big_foo([3, 3], [3, 1])

    assert big_foo([5, 3, 2], [1]) == result_a
    assert big_foo([3, 3], [3, 1]) == result_b

    cache.delete_memoized(big_foo, [5, 3, 2], [1])

    assert big_foo([5, 3, 2], [1]) != result_a
    assert big_foo([3, 3], [3, 1]) == result_b

    # Cleanup bigfoo 5,1 5,2 or it might conflict with
    # following run if it also uses memecache
    cache.delete_memoized(big_foo, [5, 3, 2], [1])
    cache.delete_memoized(big_foo, [3, 3], [1])


def test_memoize_kwargs(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b=None):
        return a + sum(b.values()) + random.randrange(0, 100000)

    result_a = big_foo(1, dict(one=1, two=2))
    result_b = big_foo(5, dict(three=3, four=4))

    assert big_foo(1, dict(one=1, two=2)) == result_a
    assert big_foo(5, dict(three=3, four=4)) == result_b

    cache.delete_memoized(big_foo, 1, dict(one=1, two=2))

    assert big_foo(1, dict(one=1, two=2)) != result_a
    assert big_foo(5, dict(three=3, four=4)) == result_b


def test_memoize_kwargonly(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a=None):
        if a is None:
            a = 0
        return a + random.random()

    result_a = big_foo()
    result_b = big_foo(5)

    assert big_foo() == result_a
    assert big_foo() < 1
    assert big_foo(5) == result_b
    assert big_foo(5) >= 5 and big_foo(5) < 6


def test_memoize_arg_kwarg(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def f(a, b, c=1):
        return a + b + c + random.randrange(0, 100000)

    assert f(1, 2) == f(1, 2, c=1)
    assert f(1, 2) == f(1, 2, 1)
    assert f(1, 2) == f(1, 2)
    assert f(1, 2, 3) != f(1, 2)

    with pytest.raises(TypeError):
        f(1)


def test_memoize_arg_kwarg_var_keyword(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def f(a, b, c=1, **kwargs):
        return a + b + c + random.randrange(0, 100000) + sum(list(kwargs.values()))

    assert f(1, 2) == f(1, 2, c=1)
    assert f(1, 2) == f(1, 2, 1)
    assert f(1, 2) == f(1, 2)
    assert f(1, 2, d=5, e=8) == f(1, 2, e=8, d=5)
    assert f(1, b=2, c=3, d=5, e=8) == f(1, 2, e=8, d=5, b=2, c=3)
    assert f(1, 2, 3) != f(1, 2)
    assert f(1, 2, 3) != f(1, 2)

    with pytest.raises(TypeError):
        f(1)


def test_memoize_classarg(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def bar(a):
        return a.value + random.random()

    class Adder(object):
        def __init__(self, value):
            self.value = value

    adder = Adder(15)
    adder2 = Adder(20)

    y = bar(adder)
    z = bar(adder2)

    assert y != z
    assert bar(adder) == y
    assert bar(adder) != z
    adder.value = 14
    assert bar(adder) == y
    assert bar(adder) != z

    assert bar(adder) != bar(adder2)
    assert bar(adder2) == z


def test_memoize_classfunc(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    class Adder(object):
        def __init__(self, initial):
            self.initial = initial

        @cache.memoize()
        def add(self, b):
            return self.initial + b

    adder1 = Adder(1)
    adder2 = Adder(2)

    x = adder1.add(3)
    assert adder1.add(3) == x
    assert adder1.add(4) != x
    assert adder1.add(3) != adder2.add(3)


def test_memoize_classfunc_repr(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    class Adder(object):
        def __init__(self, initial):
            self.initial = initial

        @cache.memoize()
        def add(self, b):
            return self.initial + b

        def __repr__(self):
            return "42"

        def __caching_id__(self):
            return self.initial

    adder1 = Adder(1)
    adder2 = Adder(2)

    x = adder1.add(3)
    assert adder1.add(3) == x
    assert adder1.add(4) != x
    assert adder1.add(3) != adder2.add(3)


def test_memoize_classfunc_delete(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    # skip the test for filesystem as there is no delete there
    if cache.config['CACHE_TYPE'] in ['filesystem']:
        return

    class Adder(object):
        def __init__(self, initial):
            self.initial = initial

        @cache.memoize()
        def add(self, b):
            return self.initial + b + random.random()

    adder1 = Adder(1)
    adder2 = Adder(2)

    a1 = adder1.add(3)
    a2 = adder2.add(3)

    assert a1 != a2
    assert adder1.add(3) == a1
    assert adder2.add(3) == a2

    cache.delete_memoized(adder1.add)

    a3 = adder1.add(3)
    a4 = adder2.add(3)

    assert not a1 == a3
    # self.assertNotEqual(a1, a3)

    assert a1 != a3

    assert a2 == a4
    # self.assertEqual(a2, a4)

    cache.delete_memoized(Adder.add)

    a5 = adder1.add(3)
    a6 = adder2.add(3)

    assert not a5 == a6
    assert not a3 == a5
    assert not a4 == a6


def test_memoize_classmethod_delete(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    class Mock(object):
        @classmethod
        @cache.memoize(5)
        def big_foo(cls, a, b):
            return a + b + random.randrange(0, 100000)

    result = Mock.big_foo(5, 2)
    result2 = Mock.big_foo(5, 3)

    time.sleep(1)

    assert Mock.big_foo(5, 2) == result
    assert Mock.big_foo(5, 2) == result
    assert Mock.big_foo(5, 3) != result
    assert Mock.big_foo(5, 3) == result2

    cache.delete_memoized(Mock.big_foo)

    assert Mock.big_foo(5, 2) != result
    assert Mock.big_foo(5, 3) != result2


def test_memoize_classmethod_delete_with_args(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return
    if cache.config['CACHE_TYPE'] in ['redis']:
        return

    class Mock(object):
        @classmethod
        @cache.memoize(5)
        def big_foo(cls, a, b):
            return a + b + random.randrange(0, 100000)

    result = Mock.big_foo(5, 2)
    result2 = Mock.big_foo(5, 3)

    time.sleep(1)

    assert Mock.big_foo(5, 2) == result
    assert Mock.big_foo(5, 2) == result
    assert Mock.big_foo(5, 3) != result
    assert Mock.big_foo(5, 3) == result2

    with pytest.raises(ValueError):
        cache.delete_memoized(Mock.big_foo, 5, 2)

    assert Mock.big_foo(5, 2) == result
    assert Mock.big_foo(5, 3) == result2

    cache.delete_memoized(Mock.big_foo, Mock, 5, 2)

    assert Mock.big_foo(5, 2) != result
    assert Mock.big_foo(5, 3) == result2


def test_memoize_forced_update(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    forced_update = False

    @cache.memoize(5, forced_update=lambda: forced_update)
    def big_foo(a, b):
        return a + b + random.randrange(0, 100000)

    result = big_foo(5, 2)
    time.sleep(1)
    assert big_foo(5, 2) == result

    forced_update = True
    new_result = big_foo(5, 2)
    assert new_result != result

    forced_update = False
    time.sleep(1)
    assert big_foo(5, 2) == new_result


def test_memoize_forced_update_parameters(app, caches):
    from collections import Counter

    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    call_counter = Counter()
    call_params = {}
    forced_update = False

    def forced_update_func(a, b):
        call_counter[1] += 1
        call_params[call_counter[1] - 1] = (a, b)

        return forced_update

    @cache.memoize(5, forced_update=forced_update_func)
    def memoized_func(a, b):
        return a + b + random.randrange(0, 100000)

    # Save the value for later inspection
    result = memoized_func(5, 2)
    # forced_update_func should have been called twice; once by memoize itself, once by
    # _memoize_version…
    assert call_counter[1] == 2
    # …with the values we called the function with
    assert call_params[0] == (5, 2)
    assert call_params[1] == (5, 2)
    time.sleep(1)

    # Calling the function again should return the cached value
    assert memoized_func(5, 2) == result
    # forced_update_func should have been called two more times…
    assert call_counter[1] == 4
    # …with the values we called the function with
    assert call_params[2] == (5, 2)
    assert call_params[3] == (5, 2)

    # Tell forced_update_func to return True next time
    forced_update = True
    # Save the new result…
    new_result = memoized_func(5, 2)
    # …which, due to the random number in the function, should be different from the old one
    assert new_result != result
    # forced_update_func should have been called two more times again…
    assert call_counter[1] == 6
    # …with the values we called the function with
    assert call_params[4] == (5, 2)
    assert call_params[5] == (5, 2)

    # Now stop forced updating again
    forced_update = False
    time.sleep(1)
    # The function should return the same value as it did last time
    assert memoized_func(5, 2) == new_result
    # forced_update_func should have been called two more times again…
    assert call_counter[1] == 8
    # …with the values we called the function with
    assert call_params[6] == (5, 2)
    assert call_params[7] == (5, 2)


def test_memoize_multiple_arg_kwarg_calls(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b, c=[1, 1], d=[1, 1]):
        return sum(a) + sum(b) + sum(c) + sum(d) + random.randrange(0, 100000)  # noqa

    result_a = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])

    assert big_foo([5, 3, 2], [1], d=[3, 3], c=[3, 3]) == result_a
    assert big_foo(b=[1], a=[5, 3, 2], c=[3, 3], d=[3, 3]) == result_a
    assert big_foo([5, 3, 2], [1], [3, 3], [3, 3]) == result_a


def test_memoize_multiple_arg_kwarg_delete(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b, c=[1, 1], d=[1, 1]):
        return sum(a) + sum(b) + sum(c) + sum(d) + random.randrange(0, 100000)  # noqa

    result_a = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])
    cache.delete_memoized(big_foo, [5, 3, 2], [1], [3, 3], [3, 3])
    result_b = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])
    assert result_a != result_b

    cache.delete_memoized(big_foo, [5, 3, 2], b=[1], c=[3, 3], d=[3, 3])
    result_b = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])
    assert result_a != result_b

    cache.delete_memoized(big_foo, [5, 3, 2], [1], c=[3, 3], d=[3, 3])
    result_a = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])
    assert result_a != result_b

    cache.delete_memoized(big_foo, [5, 3, 2], b=[1], c=[3, 3], d=[3, 3])
    result_a = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])
    assert result_a != result_b

    cache.delete_memoized(big_foo, [5, 3, 2], [1], c=[3, 3], d=[3, 3])
    result_b = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])
    assert result_a != result_b

    cache.delete_memoized(big_foo, [5, 3, 2], [1], [3, 3], [3, 3])
    result_a = big_foo([5, 3, 2], [1], c=[3, 3], d=[3, 3])
    assert result_a != result_b


def test_memoize_kwargs_to_args(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    def big_foo(a, b, c=None, d=None):
        return sum(a) + sum(b) + random.randrange(0, 100000)

    expected = (1, 2, 'foo', 'bar')

    args, kwargs = cache._memoize_kwargs_to_args(big_foo, 1, 2, 'foo', 'bar')
    assert (args == expected)
    args, kwargs = cache._memoize_kwargs_to_args(big_foo, 2, 'foo', 'bar', a=1)
    assert (args == expected)
    args, kwargs = cache._memoize_kwargs_to_args(big_foo, a=1, b=2, c='foo', d='bar')
    assert (args == expected)
    args, kwargs = cache._memoize_kwargs_to_args(big_foo, d='bar', b=2, a=1, c='foo')
    assert (args == expected)
    args, kwargs = cache._memoize_kwargs_to_args(big_foo, 1, 2, d='bar', c='foo')
    assert (args == expected)


def test_memoize_when_using_args_unpacking(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(*args):
        return sum(args) + random.randrange(0, 100000)

    result_a = big_foo(1, 2)
    result_b = big_foo(1, 3)

    assert big_foo(1, 2) == result_a
    assert big_foo(1, 3) == result_b
    assert big_foo(1, 2) != result_b
    assert big_foo(1, 3) != result_a

    cache.delete_memoized(big_foo)

    assert big_foo(1, 2) != result_a
    assert big_foo(1, 3) != result_b


def test_memoize_when_using_variable_mix_args_unpacking(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    @cache.memoize()
    def big_foo(a, b, *args, **kwargs):
        return sum([a, b]) + sum(args) + sum(kwargs.values()) + random.randrange(0, 100000)

    result_a = big_foo(1, 2, 3, 4, x=2, y=5)
    result_b = big_foo(4, 7, 7, 2, x=1, y=4)

    assert big_foo(1, 2, 3, 4, x=2, y=5) == result_a
    assert big_foo(4, 7, 7, 2, x=1, y=4) == result_b
    assert big_foo(1, 2, 3, 4, x=2, y=5) != result_b
    assert big_foo(4, 7, 7, 2, x=1, y=4) != result_a

    cache.delete_memoized(big_foo)

    assert big_foo(1, 2, 3, 4, x=2, y=5) != result_a
    assert big_foo(4, 7, 7, 2, x=1, y=4) != result_b


def test_memoize_none(app, caches):
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return
    if cache.config['CACHE_TYPE'] in ['redissentinel', 'redis']:
        return

    from collections import Counter

    call_counter = Counter()

    @cache.memoize(cache_none=True)
    def memoize_none(param):
        call_counter[param] += 1

        return None

    memoize_none(1)

    # The memoized function should have been called
    assert call_counter[1] == 1

    # Next time we call the function, the value should be coming from the cache…
    assert memoize_none(1) is None

    # …thus, the call counter should remain 1
    assert call_counter[1] == 1

    cache.clear()

    memoize_none(1)
    assert call_counter[1] == 2


def test_memoize_never_accept_none(app, caches):
    """Asserting that when cache_none is False, we always
       assume a None value returned from .get() means the key is not found
    """
    from collections import Counter

    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # skip the test for memcached
    if cache.config['CACHE_TYPE'] in ['memcached', 'gaememcached', 'saslmemcached', 'spreadsaslmemcached']:
        return

    call_counter = Counter()

    @cache.memoize()
    def memoize_none(param):
        call_counter[param] += 1

        return None

    memoize_none(1)

    # The memoized function should have been called
    assert call_counter[1] == 1

    # Next time we call the function, the value should be coming from the cache…
    # But the value is None and so we treat it as uncached.
    assert memoize_none(1) is None

    # …thus, the call counter should increment to 2
    assert call_counter[1] == 2

    cache.clear()

    memoize_none(1)
    assert call_counter[1] == 3



from falcon import HTTP_200, COMBINED_METHODS, __version__ as FALCONVERSION
from falcon_caching.options import CacheEvictionStrategy, HttpMethods
import logging
import re
import msgpack
from typing import TYPE_CHECKING, Any, Dict, Tuple

if TYPE_CHECKING:
    from falcon_caching.cache import Cache

logger = logging.getLogger(__name__)

_DECORABLE_METHOD_NAME = re.compile(r'^on_({})(_\w+)?$'.format(
    '|'.join(method.lower() for method in COMBINED_METHODS)))

# what is the Falcon main version (eg 2 or 3, etc)
FALCONVERSION_MAIN = int(FALCONVERSION.split('.')[0])


class Middleware:
    """ It integrates a cache object with Falcon by turning it into
    a Falcon Middleware
    """

    def __init__(self, cache: 'Cache', config: Dict[str, Any]) -> None:
        self.cache = cache
        self.cache_config = config

    def process_resource(self, req, resp, resource, params):
        """ Determine if the given request is marked for caching and if yes,
        then look it up in the cache and if found, then return the cached value
        """

        # Step 1: for 'rest-based' and 'rest&time-based' eviction strategies the
        # POST/PATCH/PUT/DELETE calls are never cached, they should never be
        # loaded from cache as they must always execute,
        # so for those we don't need to try to search the cache
        if self.cache_config['CACHE_EVICTION_STRATEGY'] in [CacheEvictionStrategy.rest_based,
                                                            CacheEvictionStrategy.rest_and_time_based] \
            and req.method.upper() in [HttpMethods.POST,
                                       HttpMethods.PATCH,
                                       HttpMethods.PUT,
                                       HttpMethods.DELETE]:
            return

        # Step 2: determine whether the given responder has caching setup
        # and if not then short-circuit to save on the lookup of request in the cache
        # as anyhow this request was not marked to be cached

        # find out which responder ("on_..." method) is going to be used to process this request
        responder = None
        for _method in dir(resource):
            if _DECORABLE_METHOD_NAME.match(_method) and _method[3:].upper() == req.method.upper():
                responder = _method
                break

        if responder:
            # get the name of the responder wrapper, which for cached objects is 'cache_wrap'
            # see the "Cache.cache" decorator in cache.py
            responder_wrapper_name = getattr(getattr(resource, responder), '__name__')

            # is the given method (or its class) decorated by the cache_wrap being the topmost decorator?
            if responder_wrapper_name == 'cache_wrap':
                logger.debug(" This endpoint is decorated by 'cache' being the topmost decorator.")
            else:
                # 'cache_wrap' is not the topmost decorator - let's check whether 'cache' is
                # any of the other decorator on this method (not the topmost):
                # this requires the use of @register(decor1, decor2) as the decorator
                if hasattr(getattr(resource, responder), '_decorators') and \
                        'cache' in [d._decorator_name for d in getattr(resource, responder)._decorators
                                    if hasattr(d, '_decorator_name')]:
                    logger.debug(" This endpoint is decorated by 'cache', but it is NOT the topmost decorator.")
                else:
                    # no cache was requested on this responder as no decorator at all
                    logger.debug(" No 'cache' was requested for this endpoint.")
                    return

        # Step 3: look up the record in the cache
        key = self.generate_cache_key(req)
        data = self.cache.get(key)

        if data:
            # if the CACHE_CONTENT_TYPE_JSON_ONLY = True, then we are NOT
            # caching the response's Content-Type, only its body
            if self.cache_config['CACHE_CONTENT_TYPE_JSON_ONLY']:
                if FALCONVERSION_MAIN < 3:
                    resp.body = self.deserialize(data)
                else:
                    resp.text = self.deserialize(data)
            else:
                if FALCONVERSION_MAIN < 3:
                    resp.content_type, resp.body = self.deserialize(data)
                else:
                    resp.content_type, resp.text = self.deserialize(data)
            resp.status = HTTP_200
            req.context.cached = True

            # Short-circuit any further processing to skip any remaining
            # 'process_request' and 'process_resource' methods, as well as
            # the 'responder' method that the request would have been routed to.
            # However, any 'process_response' middleware methods will still be called.
            resp.complete = True

    def process_response(self, req, resp, resource, req_succeeded):
        """ Cache the response if this request qualifies and has not been cached yet
        or for rest-based and rest-and-time-based evict the record from the cache if
        the request method is POST/PATCH/PUT or DELETE """

        # Step 1: for 'rest-based' and 'rest&time-based' eviction strategies the
        # POST/PATCH/PUT/DELETE calls are never cached and even more they
        # invalidate the record cached by the GET method
        if self.cache_config['CACHE_EVICTION_STRATEGY'] in [CacheEvictionStrategy.rest_based,
                                                            CacheEvictionStrategy.rest_and_time_based] \
            and req.method.upper() in [HttpMethods.POST,
                                       HttpMethods.PATCH,
                                       HttpMethods.PUT,
                                       HttpMethods.DELETE]:
            # get the cache key created by the GET method (assuming there was one)
            key = self.generate_cache_key(req, method='GET')
            self.cache.delete(key)
            return

        # Step 2: if it is marked to be cached, but has not yet been cached
        # then we cache it
        if hasattr(req.context, 'cache') and req.context.cache \
                and (not hasattr(req.context, 'cached') or not req.context.cached):
            key = self.generate_cache_key(req)
            value = self.serialize(req, resp, resource)

            # for the REST-based strategy there is no timeout, the cached record never expires
            if self.cache_config['CACHE_EVICTION_STRATEGY'] in [CacheEvictionStrategy.rest_based]:
                # timeout 0 - never expires
                timeout = 0
            else:
                # for the time-based and rest-and-time-based eviction strategy the
                # cached record expires
                timeout = req.context.cache_timeout if hasattr(req.context, 'cache_timeout') else 600

            self.cache.set(key, value, timeout=timeout)

    @staticmethod
    def generate_cache_key(req, method: str = None) -> str:
        """ Generate the cache key from the request using the path and the method """

        path = req.path
        if path.endswith('/'):
            path = path[:-1]

        if not method:
            method = req.method

        return f'{path}:{method.upper()}'

    def serialize(self, req, resp, resource) -> bytes:
        """ Serializes the response, so it can be cached.

        If CACHE_CONTENT_TYPE_JSON_ONLY = False (default), then we need to
        keep the response Content-Type header, so we need to serialize the response
        body with the content type with msgpack, which takes away performance.

        For this reason the user can set CACHE_CONTENT_TYPE_JSON_ONLY = True, in
        which case the response Content-Type is NOT cached, so it will be
        the default - which is application/json. That should be fine for most
        REST APIs and should bring a nice performance bump by avoiding the msgpack
        serialization.
        """
        if self.cache_config['CACHE_CONTENT_TYPE_JSON_ONLY']:
            if FALCONVERSION_MAIN < 3:
                return resp.body
            else:
                return resp.text
        else:
            if FALCONVERSION_MAIN < 3:
                return msgpack.packb([resp.content_type, resp.body], use_bin_type=True)
            else:
                return msgpack.packb([resp.content_type, resp.text], use_bin_type=True)

    def deserialize(self, data: bytes) -> Tuple[str, Any]:
        """ Deserializes the cached record into the response Body
        or the Content-Type and Body
        """
        if self.cache_config['CACHE_CONTENT_TYPE_JSON_ONLY']:
            return data
        else:
            return msgpack.unpackb(data, raw=False)



""" Functional tests to test an example app with the different
cache backends and eviction strategies
"""
import json
import os
import pytest
import random
import time

from falcon import API, testing
from falcon_caching import Cache
from falcon_caching.backends.memcache import (
    MemcachedCache,
    SASLMemcachedCache,
    SpreadSASLMemcachedCache
)
from falcon_caching.backends.redis import Redis, RedisSentinel
from falcon_caching.backends.filesystem import FileSystemCache
from falcon_caching.backends.simple import SimpleCache
from tests.conftest import CACHE_EXPIRES, CACHE_BUSTING_METHODS, CACHE_TYPES,\
    EVICTION_STRATEGIES, REDIS_PORT, CACHE_THRESHOLD, FALCONVERSION_MAIN
from tests.utils import get_cache, get_cache_class, get_cache_eviction_strategy,\
    delete_from_cache


def test_app_basic(client):
    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path='/randrange_cached', method="GET")
    result1 = client.simulate_get('/randrange_cached')
    assert 100000 >= result1.json['num'] >= 0


def test_onget_cache(client):
    """ Test a properly cached GET request
    """
    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path='/randrange_cached', method="GET")

    result1 = client.simulate_get('/randrange_cached')
    # if get_cache_class(client.app) in [SimpleCache, FileSystemCache]:
    #      pytest.skip("Skipping SimpleCache")

    result2 = client.simulate_get('/randrange_cached')

    assert result1.json['num'] == result2.json['num']


@pytest.mark.parametrize("decorator_type", [
    # testing the method-based decorators, eg calling the
    # /randrange_cached_expires endpoint
    ("method-decorated"),
    # testing the class-based decorators, eg calling the
    # /randrange_class_cached_expires endpoint
    ("class-decorated")
])
def test_cache_expires(client, decorator_type):
    """ Test expiring cache with the different eviction strategies both
    for method decorators and class decorators
    """
    if decorator_type == 'method-decorated':
        path = "/randrange_cached_expires"
    elif decorator_type == 'class-decorated':
        path = "/randrange_class_cached_expires"

    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path=path, method="GET")

    result1 = client.simulate_get(path)
    result2 = client.simulate_get(path)
    assert result1.json['num'] == result2.json['num']

    # if we wait long enough the cache expires so a new request will bring
    # a new value
    time.sleep(CACHE_EXPIRES)
    result3 = client.simulate_get(path)

    # if the eviction strategy is time-based or rest-and-time-based
    # then the cache should have already expired, so the request
    # should have gotten executed again
    eviction_strategy = get_cache_eviction_strategy(client.app)
    if eviction_strategy in ['time-based', 'rest-and-time-based']:
        assert result2.json['num'] != result3.json['num']
    elif eviction_strategy in ['rest-based']:
        assert result2.json['num'] == result3.json['num']

        # a POST command should evict the cache for rest-based
        client.simulate_post(path)

        result4 = client.simulate_get(path)
        assert result3.json['num'] != result4.json['num']


def test_rest_action_expires_cache(client):
    """ Test how cache records expires after different method requests
    with the different eviction strategies
    """
    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path="/randrange_cached", method="GET")
    for method in CACHE_BUSTING_METHODS:
        delete_from_cache(app=client.app, path="/randrange_cached", method=method)

    # normal get should chache for all eviction strategies
    result1 = client.simulate_get('/randrange_cached')
    result2 = client.simulate_get('/randrange_cached')
    assert result1.json['num'] == result2.json['num']

    eviction_strategy = get_cache_eviction_strategy(client.app)

    # if the eviction strategy is time-based then no request with
    # other method for the same endpoint should change that
    # but for the rest-based and rest-and-time-based it does change it
    if eviction_strategy == 'time-based':
        for method in CACHE_BUSTING_METHODS:
            result3 = client.simulate_get('/randrange_cached')

            # make the simulate_post(), simulate_put(), etc call:
            getattr(client, f"simulate_{method.lower()}")('/randrange_cached')

            result4 = client.simulate_get('/randrange_cached')
            assert result3.json['num'] == result4.json['num']

    elif eviction_strategy in ['rest-based', 'rest-and-time-based']:
        for method in CACHE_BUSTING_METHODS:
            result3 = client.simulate_get('/randrange_cached')

            # make the simulate_post(), simulate_put(), etc call:
            getattr(client, f"simulate_{method.lower()}")('/randrange_cached')

            result4 = client.simulate_get('/randrange_cached')
            assert result3.json['num'] != result4.json['num']


def test_explicit_caching(caches):
    """ Testing of explicit caching and retrieving of records
    """
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # delete the 'foo21' record to be sure that it is not left in the cache
    cache.delete_many("foo21", "foo22", "foo23", "foo24", "foo25")

    assert cache.has("foo21") is False

    cache.set("foo21", "bar")
    assert cache.has("foo21") is True

    # add() does not overwrite existing keys
    cache.add("foo21", "bar2")
    assert cache.get("foo21") == 'bar'

    # for SpreadSASLMemcachedCache the add() doesn't work
    if cache.cache.__class__ not in [SpreadSASLMemcachedCache]:
        cache.add("foo22", "bar2")
    else:
        cache.set("foo22", "bar2")
    assert cache.get("foo22") == 'bar2'

    # for SpreadSASLMemcachedCache the add() doesn't work
    if cache.cache.__class__ not in [SpreadSASLMemcachedCache]:
        assert cache.get_many("foo21", "foo22") == ["bar", "bar2"]
        assert cache.get_dict("foo21", "foo22") == {"foo21": "bar", "foo22": "bar2"}

    cache.set_many({"foo23": "bar3", "foo24": "bar4"})
    assert cache.get_many("foo23", "foo24") == ["bar3", "bar4"]

    # for SpreadSASLMemcachedCache the add() doesn't work
    if cache.cache.__class__ not in [SpreadSASLMemcachedCache]:
        cache.delete_many("foo21", "foo22")
        assert cache.has("foo21") is False
        assert cache.has("foo22") is False

        assert cache.has("foo23") is True
        assert cache.has("foo24") is True
        cache.clear()
        assert cache.has("foo23") is False
        assert cache.has("foo24") is False

    # these only work for Redis
    if cache.cache.__class__ in [Redis, RedisSentinel]:
        cache.set("foo25", 1)
        cache.inc("foo25")
        assert cache.get("foo25") == 2

        cache.dec("foo25")
        assert cache.get("foo25") == 1


@pytest.mark.parametrize("eviction_strategy", [
    *EVICTION_STRATEGIES
])
def test_caching_content_type(caches, eviction_strategy):
    """ Testing that the Content-Type header gets cached
    even when it is not the default, but it is set in the responder
    """
    # get the cache for the given eviction strategy
    cache = caches[eviction_strategy]

    # a resource where a custom response Cache-Type is set
    class CachedResource:
        @cache.cached(timeout=1)
        def on_get(self, req, resp):
            resp.content_type = 'mycustom/verycustom'
            if FALCONVERSION_MAIN < 3:
                resp.body = json.dumps({'num': random.randrange(0, 100000)})
            else:
                resp.text = json.dumps({'num': random.randrange(0, 100000)})

    app = API(middleware=cache.middleware)
    app.add_route('/randrange_cached', CachedResource())

    client = testing.TestClient(app)

    # the first call will cache it
    result1 = client.simulate_get('/randrange_cached')
    assert result1.headers['Content-Type'] == 'mycustom/verycustom'

    # the second call returns it from cache - but it still carries
    # the same content-type
    result2 = client.simulate_get('/randrange_cached')
    assert result1.json['num'] == result2.json['num']
    assert result2.headers['Content-Type'] == 'mycustom/verycustom'


@pytest.mark.parametrize("cache_type", [
    *CACHE_TYPES
])
@pytest.mark.parametrize("eviction_strategy", [
    *EVICTION_STRATEGIES
])
def test_caching_content_type_json_only(tmp_path, redis_server, redis_sentinel_server,
                                        memcache_server, cache_type, eviction_strategy):
    """ Testing that the Content-Type header does NOT get cached
    when the CACHE_CONTENT_TYPE_JSON_ONLY = True is set, which means
    that no msgpack serialization is used in this case, which should mean this
    gives a nice performance bump when the Content-Type header caching is
    not required because the application/json Content-Type (which is the default
    in Falcon) is sufficient for the app.
    """
    if cache_type == 'redissentinel' and os.getenv('TRAVIS', 'no') == 'yes':
        pytest.skip("Unfortunately on Travis Redis Sentinel currently can't be installed")

    # uwsgi tests should only run if running under uwsgi
    if cache_type == 'uwsgi':
        try:
            import uwsgi
        except ImportError:
            pytest.skip("uWSGI could not be imported, are you running under uWSGI?")
            return None

    if True:
        cache = Cache(
            config={
                'CACHE_EVICTION_STRATEGY': eviction_strategy,
                'CACHE_TYPE': cache_type,
                'CACHE_THRESHOLD': CACHE_THRESHOLD,
                'CACHE_DIR': tmp_path if cache_type == 'filesystem' else None,
                'CACHE_REDIS_PORT': REDIS_PORT,
                'CACHE_CONTENT_TYPE_JSON_ONLY': True  # this is what we are testing here!
            }
        )

        # a resource where a custom response Cache-Type is set
        class CachedResource:
            @cache.cached(timeout=1)
            def on_get(self, req, resp):
                resp.content_type = 'mycustom/verycustom'
                if FALCONVERSION_MAIN < 3:
                    resp.body = json.dumps({'num': random.randrange(0, 100000)})
                else:
                    resp.text = json.dumps({'num': random.randrange(0, 100000)})

        app = API(middleware=cache.middleware)
        app.add_route('/randrange_cached', CachedResource())

        client = testing.TestClient(app)

        # before making the first call let's ensure that the cache is empty
        if cache_type == 'memcached':
            cache.cache.delete("/randrange_cached:GET")
        else:
            cache.clear()

        # the first call will cache it
        result1 = client.simulate_get('/randrange_cached')
        assert result1.headers['Content-Type'] == 'mycustom/verycustom'

        # the second call returns it from cache - but as the content-type
        # is NOT cached, it will return the default 'application/json' type
        result2 = client.simulate_get('/randrange_cached')
        assert result1.json['num'] == result2.json['num']
        assert result2.headers['Content-Type'] == 'application/json'
